// Copyright (c) 1994 James Clark
// See the file COPYING for copying permission.

#ifdef __GNUG__
#pragma implementation
#endif

#include "config.h"
#include "Catalog.H"
#include "EntityManager.H"
#include "Location.H"
#include "StringOf.H"
#include "rtti.H"
#include "macros.H"
#include "CString.H"
#include "Entity.H"
#include "ExternalId.H"
#include "Message.H"
#include "StringOf.H"
#include "Vector.H"
#include "setAll.H"
#include "Boolean.H"
#include "types.H"
#include "OffsetOrderedList.H"
#include "types.H"
#include "InputSource.H"
#include "StringOf.H"
#include "StorageManager.H"
#include "GrowableVectorD.H"
#include "DOwner.H"
#include "Message.H"
#include "MessageArg.H"
#include "CharsetInfo.H"
#include "CodingSystem.H"
#include "StorageObjectInfo.H"
#include "RegisteredCodingSystem.H"
#include "constant.H"
#include "EntityManagerMessages.H"
#include "StorageObjectPosition.H"

#include <stddef.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

extern "C" {
  void *memmove(void *, const void *, size_t);
}

class ExternalInputSource;

class EntityManagerImpl : public ExternalEntityOpener {
public:
  EntityManagerImpl(StorageManager *defaultStorageManager,
		    const InputCodingSystem *defaultCodingSystem,
		    const UnivCharsetDesc &catalogCharset);
  void registerStorageManager(StorageManager *);
  void registerCodingSystem(const char *, const InputCodingSystem *);
  InputSource *openExternalEntity(const ExternalEntity &,
				  EntityOrigin *,
				  const CharsetInfo &,
				  const SubstTable<Char> *,
				  const CString &,
				  Boolean mayRewind,
				  Messenger &);
  InputSource *openDefaultSgmlDecl(Messenger &);
  InputSource *openDocumentEntity(const CString *sysids,
				  size_t nSysids,
				  const UnivCharsetDesc &charsetDesc,
				  Messenger &);
  StorageManager *lookupStorageType(const CString &, const CharsetInfo &);
  StorageManager *lookupStorageType(const char *);
  
  const InputCodingSystem *lookupCodingSystem(const CString &,
					      const CharsetInfo &,
					      const char *&);
  const InputCodingSystem *lookupCodingSystem(const char *);
  void addCatalogSysid(const CString &sysid, Boolean mustExist);
  Boolean getExternalEntityStorage(const ExternalEntity &entity,
				   const CharsetInfo &docCharset,
				   const SubstTable<Char> *subst,
				   const CString &pero,
				   Messenger &,
				   GrowableVectorD<StorageObjectSpec> &specs);
private:
  EntityManagerImpl(const EntityManagerImpl &); // undefined
  void operator=(const EntityManagerImpl &); // undefined
  static const StorageObjectSpec *defStorageObject(const Location &);
  void loadCatalog(Messenger &);
  Boolean resolveSystemId(const CString &str,
			  const Location &defLocation,
			  const CharsetInfo &idCharset,
			  Messenger &mgr,
			  GrowableVectorD<StorageObjectSpec> &specs);
  Boolean parseSystemId(const Char *,
			size_t,
			const CharsetInfo &docCharset,
			Messenger &,
			GrowableVectorD<StorageObjectSpec> &specs);
  Boolean parseConcat(const Char *s,
		      size_t n,
		      const CharsetInfo &docCharset,
		      Messenger &,
		      GrowableVectorD<StorageObjectSpec> &specs);
  ExternalInputSource *
    makeInputSource(GrowableVectorD<StorageObjectSpec> &specs,
		    EntityOrigin *origin,
		    Boolean mayRewind,
		    Messenger &);
  Boolean convertId(CString &id,
		    const CharsetInfo &fromCharset,
		    const StorageManager *sm,
		    Messenger &);
  static Boolean matchKey(const CString &type, const char *s,
			  const CharsetInfo &docCharset);
  Boolean loadedCatalog_;
  Catalog catalog_;
  GrowableVectorD<CString> catalogSysids_;
  GrowableVector<PackedBoolean> catalogSysidMustExist_;
  CharsetInfo catalogCharset_;
  GrowableVector<StorageManager *> storageManagers_;
  GrowableVector<RegisteredCodingSystem> codingSystems_;
  StorageManager *defaultStorageManager_;
  const InputCodingSystem *defaultCodingSystem_;
};

class ExternalInfoImpl : public ExternalInfo {
  RTTI_CLASS
public:
  ExternalInfoImpl(GrowableVectorD<StorageObjectSpec> &specs);
  const StorageObjectSpec &spec(size_t i) const;
  size_t nSpecs() const;
  void noteRS(Offset);
  void noteStorageObjectEnd(Offset);
  void noteInsertedRSs();
  void setDecoder(size_t i, Decoder *);
  Boolean convertOffset(Offset, StorageObjectLocation &) const;
private:
  GrowableVectorD<StorageObjectSpec> specs_;
  Vector<StorageObjectPosition> position_;
  size_t currentIndex_;
  // list of inserted RSs
  OffsetOrderedList rsList_;
  Boolean huge_;
};

class ExternalInputSource : public InputSource {
public:
  ExternalInputSource(GrowableVectorD<StorageObjectSpec> &specs,
		      Vector<StorageObjectInfo> &sov,
		      EntityOrigin *origin,
		      Boolean mayRewind);
  void pushCharRef(Char, const NamedCharRef &);
  ~ExternalInputSource();
private:
  Xchar fill(Messenger &);
  Boolean rewind(Messenger &);
  void willNotRewind();

  void init();
  void noteRS();
  void noteRSAt(const Char *);
  void reallocateBuffer(size_t size);
  void insertChar(Char);
  static const Char *findNext(Char ch, const Char *start, const Char *end);
  static const Char *findNextCrOrLf(const Char *start, const Char *end);

  ExternalInfoImpl *info_;
  Char *buf_;
  const Char *bufLim_;
  Offset bufLimOffset_;
  size_t bufSize_;
  size_t readSize_;
  Vector<StorageObjectInfo> sov_;
  StorageObject *so_;
  size_t soIndex_;
  Boolean insertRS_;
  Decoder *decoder_;
  int minBytesPerChar_;
  const char *leftOver_;
  size_t nLeftOver_;
  Boolean mayRewind_;
  enum RecordType {
    unknown,
    crUnknown,
    crlf,
    lf,
    cr,
    raw
    };
  RecordType recordType_;
};

const Char RS = '\n';
const Char RE = '\r';
const char lineEnd = '\n';

EntityManager::EntityManager()
: impl_(0)
{
}

EntityManager::EntityManager(StorageManager *defaultStorageManager,
			     const InputCodingSystem *defaultCodingSystem,
			     const UnivCharsetDesc &catalogCharset)
{
  impl_ = new EntityManagerImpl(defaultStorageManager,
				defaultCodingSystem,
				catalogCharset);
}

void EntityManager::init(StorageManager *defaultStorageManager,
			 const InputCodingSystem *defaultCodingSystem,
			 const UnivCharsetDesc &catalogCharset)

{
  delete impl_;
  impl_ = new EntityManagerImpl(defaultStorageManager,
				defaultCodingSystem,
				catalogCharset);
}

EntityManager::~EntityManager()
{
  delete impl_;
}

ExternalEntityOpener *EntityManager::externalEntityOpener()
{
  return impl_;
}

InputSource *EntityManager::openDocumentEntity(const CString *sysids,
					       size_t nSysids,
					       const UnivCharsetDesc &charset,
					       Messenger &mgr)
{
  return impl_->openDocumentEntity(sysids, nSysids, charset, mgr);
}

void EntityManager::registerStorageManager(StorageManager *sm)
{
  impl_->registerStorageManager(sm);
}

void EntityManager::registerCodingSystem(const char *s,
					 const InputCodingSystem *ics)
{
  impl_->registerCodingSystem(s, ics);
}


void EntityManager::addCatalogSysid(const CString &str, Boolean mustExist)
{
  impl_->addCatalogSysid(str, mustExist);
}

Boolean
EntityManager::getExternalEntityStorage(const ExternalEntity &entity,
					const CharsetInfo &docCharset,
					const SubstTable<Char> *subst,
					const CString &pero,
					Messenger &mgr,
					GrowableVectorD<StorageObjectSpec> &specs)
{
  return impl_->getExternalEntityStorage(entity, docCharset, subst, pero,
					 mgr, specs);
}

Boolean EntityManager::externalize(const ExternalInfo *info,
				   Offset off,
				   StorageObjectLocation &loc)
{
  if (!info)
    return false;
  const ExternalInfoImpl *p = ptr_cast(ExternalInfoImpl, info);
  if (!p)
    return false;
  return p->convertOffset(off, loc);
}

EntityManagerImpl::EntityManagerImpl(StorageManager *defaultStorageManager,
				     const InputCodingSystem *defaultCodingSystem,
				     const UnivCharsetDesc &catalogCharset)
: defaultStorageManager_(defaultStorageManager),
  defaultCodingSystem_(defaultCodingSystem),
  loadedCatalog_(false),
  catalogCharset_(catalogCharset)
{
}

InputSource *
EntityManagerImpl::openDocumentEntity(const CString *sysids,
				      size_t nSysids,
				      const UnivCharsetDesc &charsetDesc,
				      Messenger &mgr)
{
  CharsetInfo charsetInfo(charsetDesc);
  GrowableVectorD<StorageObjectSpec> specs;
  for (int i = 0; i < nSysids; i++)
    if (!parseSystemId(sysids[i].pointer(),
		       sysids[i].length(),
		       charsetInfo,
		       mgr,
		       specs))
      return 0;
  return makeInputSource(specs,
			 new EntityOrigin((Entity *)0, Location(), 0),
			 1,
			 mgr);
}

ExternalInputSource *
EntityManagerImpl::makeInputSource(GrowableVectorD<StorageObjectSpec> &specs,
				   EntityOrigin *origin,
				   Boolean mayRewind,
				   Messenger &mgr)
{
  Vector<StorageObjectInfo> info;
  info.init(specs.length());

  for (size_t i = 0; i < specs.length(); i++) {
    StorageManager *sm = lookupStorageType(specs[i].storageType);

    info[i].storageObject = sm->makeStorageObject(specs[i].id, mgr);
    info[i].codingSystem = lookupCodingSystem(specs[i].codingSystem);
    if (!info[i].codingSystem)
      info[i].codingSystem = defaultCodingSystem_;
  }
  return new ExternalInputSource(specs, info, origin, mayRewind);
}


InputSource *
EntityManagerImpl::openExternalEntity(const ExternalEntity &entity,
				      EntityOrigin *origin,
				      const CharsetInfo &docCharset,
				      const SubstTable<Char> *subst,
				      const CString &pero,
				      Boolean mayRewind,
				      Messenger &mgr)
{
  GrowableVectorD<StorageObjectSpec> specs;
  
  if (!getExternalEntityStorage(entity, docCharset, subst, pero, mgr, specs))
    return 0;
  return makeInputSource(specs, origin, mayRewind, mgr);
}

Boolean
EntityManagerImpl::getExternalEntityStorage(const ExternalEntity &entity,
					    const CharsetInfo &docCharset,
					    const SubstTable<Char> *subst,
					    const CString &pero,
					    Messenger &mgr,
					    GrowableVectorD<StorageObjectSpec> &specs)
{
  const ExternalId &id = entity.externalId();
  const CString *str = 0;
  Location defLocation;
  if (!loadedCatalog_)
    loadCatalog(mgr);
  if (catalog_.override() || !id.systemIdString()) {
    Catalog::DeclType declType;
    Boolean usePname = false;
    CString pname;
    switch (entity.declType()) {
    case Entity::generalEntity:
      declType = Catalog::entityDecl;
      break;
    case Entity::parameterEntity:
      usePname = true;
      pname = pero;
      pname += entity.name();
      declType = Catalog::entityDecl;
      break;
    case Entity::doctype:
      declType = Catalog::doctypeDecl;
      break;
    case Entity::linktype:
      declType = Catalog::linktypeDecl;
      break;
    default:
      CANNOT_HAPPEN();
    }
    if (!catalog_.lookupEntity(id.publicId(),
			       usePname ? pname : entity.name(),
			       declType, subst,
			       str, defLocation))
      str = 0;
  }
  if (!str) {
    str = id.systemIdString();
    if (!str) {
      // FIXME perhaps use name (and subst) to construct filename.
      mgr.message(EntityManagerMessages::cannotGenerateSystemId,
		  StringMessageArg(entity.name()),
		  entity.defLocation());
      return false;
    }
    defLocation = entity.defLocation();
  }
  return resolveSystemId(*str, defLocation, docCharset, mgr, specs);
}

Boolean
EntityManagerImpl::resolveSystemId(const CString &str,
				   const Location &defLocation,
				   const CharsetInfo &idCharset,
				   Messenger &mgr,
				   GrowableVectorD<StorageObjectSpec> &specs)
{
  if (!parseSystemId(str.pointer(), str.length(), idCharset, mgr, specs))
    return false;
  const StorageObjectSpec *defSpec = defStorageObject(defLocation);
  if (defSpec) {
    StorageManager *sm = lookupStorageType(defSpec->storageType);
    for (size_t i = 0; i < specs.length(); i++) {
      if (specs[i].storageType = defSpec->storageType)
	sm->mergeId(specs[i].id, defSpec->id);
      if (specs[i].codingSystem == 0)
	specs[i].codingSystem = defSpec->codingSystem;
    }
  }
  return true;
}


const StorageObjectSpec *
EntityManagerImpl::defStorageObject(const Location &location)
{
  Offset off;
  const ExternalInfo *info;
  Location loc(location);
  for (;;) {
    if (loc.origin().isNull())
      return 0;
    const EntityOrigin *entityOrigin = loc.origin()->asEntityOrigin();
    if (entityOrigin) {
      NamedCharRef ref;
      Index index = loc.index();
      while (entityOrigin->lookupIndex(index, off, ref)
	     == EntityOrigin::replacementCharIndex)
	index = ref.refStartIndex();
      info = entityOrigin->externalInfo();
      if (info)
	break;
      const InternalEntity *internal = entityOrigin->entity()->asInternalEntity();
      if (!internal)
	return 0;
      loc = internal->text().charLocation(off);
    }
    else
      loc = loc.origin()->parent();
  }
  StorageObjectLocation soLoc;
  if (!EntityManager::externalize(info, off, soLoc))
    return 0;
  return soLoc.storageObjectSpec;
}


// Append onto specs

Boolean EntityManagerImpl::parseSystemId(const Char *s,
					 size_t n,
					 const CharsetInfo &docCharset,
					 Messenger &mgr,
					 GrowableVectorD<StorageObjectSpec> &specs)
{
  Char colon = docCharset.execToDesc(':');
  for (size_t i = 0; i < n; i++)
    if (s[i] == colon) {
      CString key(s, i);
      if (matchKey(key, "CONCAT", docCharset))
	return parseConcat(s + i + 1, n - i - 1, docCharset, mgr, specs);
      StorageManager *sm = lookupStorageType(key, docCharset);
      if (sm) {
	StorageObjectSpec &spec = specs.grow();
	spec.storageType = sm->type();
	spec.id.set(s + i + 1, n - i - 1);
	if (!convertId(spec.id, docCharset, sm, mgr))
	  return 0;
	return 1;
      }
      const char *codingSystemName;
      const InputCodingSystem *codingSystem
	= lookupCodingSystem(key, docCharset, codingSystemName);
      if (codingSystem) {
	size_t startLength = specs.length();
	if (!parseSystemId(s + i + 1, n - i - 1, docCharset, mgr, specs))
	  return 0;
	for (size_t j = startLength; j < specs.length(); j++)
	  if (!specs[j].codingSystem)
	    specs[j].codingSystem = codingSystemName;
	return 1;
      }
      if (matchKey(key, "HUGE", docCharset)) {
	size_t startLength = specs.length();
	if (!parseSystemId(s + i + 1, n - i - 1, docCharset, mgr, specs))
	  return 0;
	for (size_t j = startLength; j < specs.length(); j++)
	  specs[j].flags |= StorageObjectSpec::huge;
	return 1;
      }
      if (matchKey(key, "RAW", docCharset)) {
	size_t startLength = specs.length();
	if (!parseSystemId(s + i + 1, n - i - 1, docCharset, mgr, specs))
	  return 0;
	for (size_t j = startLength; j < specs.length(); j++)
	  specs[j].flags |= StorageObjectSpec::raw;
	return 1;
      }
      break;
    }
  StorageObjectSpec &spec = specs.grow();
  spec.id.set(s, n);
  if (!convertId(spec.id, docCharset, defaultStorageManager_, mgr))
    return 0;
  spec.storageType = defaultStorageManager_->type();
  return 1;
}

Boolean EntityManagerImpl::convertId(CString &id,
				     const CharsetInfo &fromCharset,
				     const StorageManager *sm,
				     Messenger &mgr)
{
  CString newId;
  for (size_t i = 0; i < id.length(); i++) {
    UnivChar univ;
    WideChar wide;
    ISet<WideChar> wideSet;
    if (!fromCharset.descToUniv(id[i], univ))
      return 0;
    if (univ == UnivCharsetDesc::rs)
      newId += sm->rsString();
    else if (univ == UnivCharsetDesc::re)
      newId += sm->reString();
    else if (sm->idCharset().univToDesc(univ, wide, wideSet) != 1
	|| wide > charMax)
      return 0;			// FIXME give error
    else
      newId += Char(wide);
  }
  id = newId;
  return 1;
}

Boolean EntityManagerImpl::parseConcat(const Char *s,
				       size_t n,
				       const CharsetInfo &docCharset,
				       Messenger &mgr,
				       GrowableVectorD<StorageObjectSpec> &specs)
{
  if (n == 0) {
    // FIXME should use location of system id for error.
    mgr.message(EntityManagerMessages::emptySystemId);
    return false;
  }
  Char plus = docCharset.execToDesc('+');
  size_t start = 0;
  for (size_t i = 0; i < n; i++)
    if (s[i] == plus) {
      if (!parseSystemId(s + start, i - start, docCharset, mgr, specs))
	return 0;
      start = i + 1;
    }
  return parseSystemId(s + start, n - start, docCharset, mgr, specs);
}

void EntityManagerImpl::addCatalogSysid(const CString &str,
					Boolean mustExist)
{
  catalogSysids_.grow() = str;
  catalogSysidMustExist_.grow() = mustExist;
}

void EntityManagerImpl::loadCatalog(Messenger &mgr)
{
  NullMessenger nullMessenger;
  loadedCatalog_ = true;
  CatalogParser parser(catalogCharset_);
  // There are unfortunately semantics associated with the way
  // the catalog divides into files, so we have to parse
  // each storage object separately.
  for (int i = 0; i < catalogSysids_.length(); i++) {
    GrowableVectorD<StorageObjectSpec> specs;
    Messenger *useMgr = catalogSysidMustExist_[i] ? &mgr : &nullMessenger;
    if (parseSystemId(catalogSysids_[i].pointer(),
		      catalogSysids_[i].length(),
		      // FIXME might be different charset for ID
		      catalogCharset_,
		      *useMgr,
		      specs)) {
      for (size_t j = 0; j < specs.length(); j++) {
	if (!catalogSysidMustExist_[i]) {
	  StorageManager *sm = lookupStorageType(specs[j].storageType);
	  DOwner<StorageObject> obj(sm->makeStorageObject(specs[j].id,
							  nullMessenger));
	  size_t dummy;
	  if (!obj || !obj->open(nullMessenger, 0, dummy))
	    continue;
	}
	GrowableVectorD<StorageObjectSpec> thisSpec(1);
	thisSpec[0] = specs[j];
	InputSource *in = makeInputSource(thisSpec,
					  new EntityOrigin((Entity *)0,
							   Location(),
							   0),
					  0,
					  mgr);
	parser.parseCatalog(&catalog_, in, mgr);
	delete in;
      }
    }
  }
}

StorageManager *
EntityManagerImpl::lookupStorageType(const CString &type,
				     const CharsetInfo &docCharset)
{
  if (type.length() == 0)
    return 0;
  if (matchKey(type, defaultStorageManager_->type(), docCharset))
    return defaultStorageManager_;
  for (int i = 0; i < storageManagers_.length(); i++)
    if (matchKey(type, storageManagers_[i]->type(), docCharset))
      return storageManagers_[i];
  return 0;
}

StorageManager *
EntityManagerImpl::lookupStorageType(const char *type)
{
  if (type == defaultStorageManager_->type())
    return defaultStorageManager_;
  for (int i = 0; i < storageManagers_.length(); i++)
    if (type == storageManagers_[i]->type())
      return storageManagers_[i];
  return 0;
}

const InputCodingSystem *
EntityManagerImpl::lookupCodingSystem(const CString &type,
				      const CharsetInfo &docCharset,
				      const char *&name)
{
  for (size_t i = 0; i < codingSystems_.length(); i++)
    if (matchKey(type, codingSystems_[i].name, docCharset)) {
      name = codingSystems_[i].name;
      return codingSystems_[i].ics;
    }
  return 0;
}

const InputCodingSystem *
EntityManagerImpl::lookupCodingSystem(const char *s)
{
  for (size_t i = 0; i < codingSystems_.length(); i++)
    if (s == codingSystems_[i].name)
      return codingSystems_[i].ics;
  return 0;
}

Boolean
EntityManagerImpl::matchKey(const CString &type,
			    const char *s,
			    const CharsetInfo &docCharset)
{
  if (strlen(s) != type.length())
    return false;
  for (int i = 0; i < type.length(); i++)
    if (docCharset.execToDesc(toupper(s[i])) != type[i]
	&& docCharset.execToDesc(tolower(s[i])) != type[i])
      return false;
  return true;
}

void EntityManagerImpl::registerStorageManager(StorageManager *sm)
{
  storageManagers_.grow() = sm;
}

void EntityManagerImpl::registerCodingSystem(const char *name,
					     const InputCodingSystem *ics)
{
  RegisteredCodingSystem &rcs = codingSystems_.grow();
  rcs.name = name;
  rcs.ics = ics;
}

InputSource *EntityManagerImpl::openDefaultSgmlDecl(Messenger &mgr)
{
  if (!loadedCatalog_)
    loadCatalog(mgr);
  const CString *sysid;
  Location loc;
  if (!catalog_.sgmlDecl(sysid, loc))
    return 0;
  GrowableVectorD<StorageObjectSpec> specs;
  if (!resolveSystemId(*sysid, loc, catalogCharset_, mgr, specs))
    return 0;
  return makeInputSource(specs,
			 new EntityOrigin((Entity *)0, Location(), 0),
			 0,
			 mgr);
}

ExternalInputSource::ExternalInputSource(GrowableVectorD<StorageObjectSpec> &specs,
					 Vector<StorageObjectInfo> &sov,
					 EntityOrigin *origin,
					 Boolean mayRewind)
: InputSource(origin, 0, 0),
  mayRewind_(mayRewind)
{
  init();
  sov.moveTo(sov_);
  info_ = new ExternalInfoImpl(specs);
  origin->setExternalInfo(info_);
}

void ExternalInputSource::init()
{
  so_ = 0;
  buf_ = 0;
  bufSize_ = 0;
  bufLim_ = 0;
  bufLimOffset_ = 0;
  insertRS_ = true;
  soIndex_ = 0;
  leftOver_ = 0;
  nLeftOver_ = 0;  
}

ExternalInputSource::~ExternalInputSource()
{
  if (buf_)
    delete [] buf_;
}

Boolean ExternalInputSource::rewind(Messenger &mgr)
{
  reset(0, 0);
  if (buf_)
    delete [] buf_;
  // reset makes a new EntityOrigin
  GrowableVectorD<StorageObjectSpec> specs(info_->nSpecs());
  size_t i;
  for (i = 0; i < specs.length(); i++)
    specs[i] = info_->spec(i);
  info_ = new ExternalInfoImpl(specs);
  entityOrigin()->setExternalInfo(info_);
  so_ = 0;
  for (i = 0; i < soIndex_; i++) {
    if (sov_[i].storageObject && !sov_[i].storageObject->rewind(mgr))
      return 0;
  }
  init();
  return 1;
}

void ExternalInputSource::willNotRewind()
{
  for (size_t i = 0; i < soIndex_; i++)
    if (sov_[i].storageObject)
      sov_[i].storageObject->willNotRewind();
  mayRewind_ = 0;
}

// Round up N so that it is a power of TO.
// TO must be a power of 2.

inline
size_t roundUp(size_t n, size_t to)
{
  return (n + (to - 1)) & ~(to - 1);
}

inline
void ExternalInputSource::noteRSAt(const Char *p)
{
  info_->noteRS(bufLimOffset_ - (bufLim_ - p));
}

inline
void ExternalInputSource::noteRS()
{
  noteRSAt(cur());
}

Xchar ExternalInputSource::fill(Messenger &mgr)
{
  ASSERT(cur() == end());
  while (end() >= bufLim_) {
    // need more data
    while (so_ == 0) {
      if (soIndex_ >= sov_.length())
	return eE;
      if (soIndex_ > 0)
	info_->noteStorageObjectEnd(bufLimOffset_ - (bufLim_ - end()));
      so_ = sov_[soIndex_].storageObject.pointer();
      if (so_) {
	decoder_ = sov_[soIndex_].codingSystem->makeDecoder();
	info_->setDecoder(soIndex_, decoder_);
	minBytesPerChar_ = sov_[soIndex_].codingSystem->minBytesPerChar();
	if (info_->spec(soIndex_).flags & StorageObjectSpec::raw) {
	  recordType_ = raw;
	  insertRS_ = false;
	}
	else
	  recordType_ = unknown;
	soIndex_++;
	// so_ might be NULL
	if (so_->open(mgr, mayRewind_, readSize_))
	  break;
	so_ = 0;
      }
      else
	soIndex_++;
    }

    size_t keepSize = end() - start();
    const size_t align = sizeof(int)/sizeof(Char);
    size_t readSizeChars = (readSize_ + (sizeof(Char) - 1))/sizeof(Char);
    readSizeChars = roundUp(readSizeChars, align);
    size_t neededSize;		// in Chars
    size_t startOffset;
    // compute neededSize and readSize
    if (nLeftOver_ == 0 && minBytesPerChar_ >= sizeof(Char)) {
      // In this case we want to do decoding in place.
      // FIXME It might be a win on some systems (Irix?) to arrange that the
      // read buffer is on a page boundary.

      if (keepSize >= size_t(-1)/sizeof(Char) - (align - 1) - insertRS_)
	abort();			// FIXME throw an exception
      
      // Now size_t(-1)/sizeof(Char) - (align - 1) - insertRS_ - keepSize > 0
      if (readSizeChars
	  > size_t(-1)/sizeof(Char) - (align - 1) - insertRS_ - keepSize)
	abort();
      neededSize = roundUp(readSizeChars + keepSize + insertRS_, align);
      startOffset = neededSize - readSizeChars - insertRS_ - keepSize;
    }
    else {
      // Needs to be room for everything before decoding.
      neededSize = (keepSize + insertRS_ + readSizeChars
		    + (nLeftOver_ + sizeof(Char) - 1)/sizeof(Char));
      // Also must be room for everything after decoding.
      size_t neededSize2
	= (keepSize + insertRS_
	   // all the converted characters
	   + (nLeftOver_ + readSize_)/minBytesPerChar_
	   // enough Chars to contain left over bytes
	   + ((readSize_ % minBytesPerChar_ + sizeof(Char) - 1)
	      / sizeof(Char)));
      if (neededSize2 > neededSize)
	neededSize = neededSize2;
      neededSize = roundUp(neededSize, align);
      if (neededSize > size_t(-1)/sizeof(Char))
	abort();
      startOffset = 0;
    }
    if (bufSize_ < neededSize)
      reallocateBuffer(neededSize);
    Char *newStart = buf_ + startOffset;
    if (newStart != start() && keepSize > 0)
      memmove(newStart, start(), keepSize*sizeof(Char));
    char *bytesStart = (char *)(buf_ + bufSize_ - readSizeChars) - nLeftOver_;
    if (nLeftOver_ > 0 && leftOver_ != bytesStart)
      memmove(bytesStart, leftOver_, nLeftOver_);
    moveStart(newStart);
    bufLim_ = end();

    size_t nread;
    if (so_->read((char *)(buf_ + bufSize_ - readSizeChars), readSize_,
		  mgr, nread)) {
      size_t nChars = decoder_->decode((Char *)end() + insertRS_,
				       bytesStart,
				       nLeftOver_ + nread,
				       &leftOver_);
      nLeftOver_ = bytesStart + nLeftOver_ + nread - leftOver_;
      if (insertRS_) {
	noteRS();
	*(Char *)end() = RS;
	advanceEnd(end() + 1);
	insertRS_ = false;
	bufLim_ += 1;
	bufLimOffset_ += 1;
      }
      bufLim_ += nChars;
      bufLimOffset_ += nChars;
      break;
    }
    else
      so_ = 0;
  }
  ASSERT(end() < bufLim_);
  if (insertRS_) {
    noteRS();
    insertChar(RS);
    insertRS_ = false;
    bufLimOffset_ += 1;
  }
  switch (recordType_) {
  case unknown:
    {
      const Char *e = findNextCrOrLf(end(), bufLim_);
      if (e) {
	if (*e == '\n') {
	  recordType_ = lf;
	  info_->noteInsertedRSs();
	  *(Char *)e = RE;
	  advanceEnd(e + 1);
	  insertRS_ = true;
	}
	else {
	  if (e + 1 < bufLim_) {
	    if (e[1] == '\n') {
	      recordType_ = crlf;
	      advanceEnd(e + 1);
	    }
	    else {
	      advanceEnd(e + 1);
	      recordType_ = cr;
	      info_->noteInsertedRSs();
	      insertRS_ = true;
	    }
	  }
	  else {
	    recordType_ = crUnknown;
	    advanceEnd(e + 1);
	  }
	}
      }
      else
	advanceEnd(bufLim_);
    }
    break;
  case crUnknown:
    {
      if (*cur() == '\n') {
	noteRS();
	advanceEnd(cur() + 1);
	recordType_ = crlf;
      }
      else {
	advanceEnd(cur() + 1);
	insertRS_ = true;
	recordType_ = cr;
	info_->noteInsertedRSs();
      }
    }
    break;
  case lf:
    {
      Char *e = (Char *)findNext('\n', end(), bufLim_);
      if (e) {
	advanceEnd(e + 1);
	*e = RE;
	insertRS_ = true;
      }
      else
	advanceEnd(bufLim_);
    }
    break;
  case cr:
    {
      const Char *e = findNext('\r', end(), bufLim_);
      if (e) {
	advanceEnd(e + 1);
	insertRS_ = true;
      }
      else
	advanceEnd(bufLim_);
    }
    break;
  case crlf:
    {
      const Char *e = end();
      for (;;) {
	e = findNext('\n', e, bufLim_);
	if (!e) {
	  advanceEnd(bufLim_);
	  break;
	}
	// Need to delete final RS if not followed by anything.
	if (e + 1 == bufLim_) {
	  bufLim_--;
	  bufLimOffset_--;
	  advanceEnd(e);
	  insertRS_ = true;
	  break;
	}
	noteRSAt(e);
	e++;
      }
    }
    break;
  case raw:
    advanceEnd(bufLim_);
    break;
  default:
    CANNOT_HAPPEN();
  }

  return nextChar();
}

const Char *ExternalInputSource::findNext(Char ch,
					  const Char *start,
					  const Char *end)
{
  for (; start < end; start++)
    if (*start == ch)
      return start;
  return 0;
}

const Char *ExternalInputSource::findNextCrOrLf(const Char *start,
						const Char *end)
{
  for (; start < end; start++)
    if (*start == '\n' || *start == '\r')
      return start;
  return 0;
}

void ExternalInputSource::pushCharRef(Char ch, const NamedCharRef &ref)
{
  ASSERT(cur() == start());
  noteCharRef(startIndex() + (cur() - start()), ref);
  insertChar(ch);
}

void ExternalInputSource::insertChar(Char ch)
{
  if (start() > buf_) {
    if (cur() > start())
      memmove((Char *)start() - 1, start(), (cur() - start())*sizeof(Char));
    moveLeft();
    *(Char *)cur() = ch;
  }
  else {
    // must have start == buf
    if (bufLim_ - buf_
	== bufSize_ - (nLeftOver_ + sizeof(Char) - 1)/sizeof(Char)) {
      if (bufSize_ == size_t(-1))
	abort();		// FIXME throw an exception
      reallocateBuffer(bufSize_ + 1);
    }
    else if (nLeftOver_ > 0 && ((char *)(bufLim_ + 1) > leftOver_)) {
      char *s = (char *)(buf_ + bufSize_) - nLeftOver_;
      memmove(s, leftOver_, nLeftOver_);
      leftOver_ = s;
    }
    if (cur() < bufLim_)
      memmove((Char *)cur() + 1, cur(), (bufLim_ - cur())*sizeof(Char));
    *(Char *)cur() = ch;
    advanceEnd(end() + 1);
    bufLim_ += 1;
  }
}

void ExternalInputSource::reallocateBuffer(size_t newSize)
{
  Char *newBuf = new Char[newSize];
  
  memcpy(newBuf, buf_, bufSize_*sizeof(Char));
  bufSize_ = newSize;
  changeBuffer(newBuf, buf_);
  bufLim_ = newBuf + (bufLim_ - buf_);
  if (nLeftOver_ > 0) {
    char *s = (char *)(newBuf + bufSize_) - nLeftOver_;
    memmove(s,
	    newBuf + (leftOver_ - (char *)buf_),
	    nLeftOver_);
    leftOver_ = s;
  }
  delete [] buf_;
  buf_ = newBuf;
}

RTTI_DEF1(ExternalInfoImpl, ExternalInfo)

ExternalInfoImpl::ExternalInfoImpl(GrowableVectorD<StorageObjectSpec> &specs)
: currentIndex_(0)
{
  specs.moveTo(specs_);
  if (specs_.length() > 0)
    huge_ = (specs_[0].flags & StorageObjectSpec::huge);
  position_.init(specs_.length());
}

void ExternalInfoImpl::setDecoder(size_t i, Decoder *decoder)
{
  position_[i].decoder = decoder;
}

void ExternalInfoImpl::noteInsertedRSs()
{
  position_[currentIndex_].insertedRSs = 1;
}

void ExternalInfoImpl::noteRS(Offset offset)
{
  if (!huge_)
    rsList_.append(offset);
  if (offset
      == (currentIndex_ == 0 ? 0 : position_[currentIndex_- 1].endOffset))
    position_[currentIndex_].startsWithRS = 1;
}

void ExternalInfoImpl::noteStorageObjectEnd(Offset offset)
{
  ASSERT(currentIndex_ < position_.length());
  // The last endOffset_ must be -1.
  if (currentIndex_ < position_.length() - 1) {
    position_[currentIndex_++].endOffset = offset;
    position_[currentIndex_].line1RS = rsList_.length();
    huge_ = (specs_[currentIndex_].flags & StorageObjectSpec::huge);
  }
}

Boolean ExternalInfoImpl::convertOffset(Offset off,
					StorageObjectLocation &ret) const
{
  ret.storageObjectSpec = 0;
  if (off == -1 || position_.length() == 0)
    return false;
  // the last endOffset_ is Offset(-1), so this will
  // terminate
  for (int i = 0; off >= position_[i].endOffset; i++)
    ;
  ret.storageObjectSpec = &specs_[i];
  Offset startOffset = i == 0 ? 0 : position_[i - 1].endOffset;
  ret.storageObjectOffset = off - startOffset;
  ret.byteIndex = ret.storageObjectOffset;
  if (specs_[i].flags & (StorageObjectSpec::huge|StorageObjectSpec::raw)) {
    ret.lineNumber = (unsigned long)-1;
    if (!(specs_[i].flags & StorageObjectSpec::raw)) {
      if (position_[i].insertedRSs)
	ret.byteIndex = (unsigned long)-1;
      else if (ret.byteIndex > 0 && position_[i].startsWithRS)
	ret.byteIndex--;	// first RS is inserted
    }
    ret.columnNumber = (unsigned long)-1;
    return true;
  }
  else {
    size_t line1RS = position_[i].line1RS;
    // line1RS is now the number of RSs that are before or on the current line.
    size_t j;
    Offset colStart;
    if (rsList_.findPreceding(off, j, colStart)) {
      if (position_[i].insertedRSs)
	ret.byteIndex -= j + 1 - line1RS;
      else if (ret.byteIndex > 0 && position_[i].startsWithRS)
	ret.byteIndex--;	// first RS is inserted
      j++;
      colStart++;
    }
    else {
      j = 0;
      colStart = 0;
    }
    // j is now the number of RSs that are before or on the current line
    // colStart is the offset of the first column
    ret.lineNumber = j - line1RS + 1 - position_[i].startsWithRS;
    // the offset of the first column
    if (colStart < startOffset)
      colStart = startOffset;
    // the RS that starts a line will be in column 0;
    // the first real character of a line will be column 1
    ret.columnNumber = 1 + off - colStart;
  }
  if (!position_[i].decoder
      || !position_[i].decoder->convertOffset(ret.byteIndex))
    ret.byteIndex = (unsigned long)-1;
  return true;
}

const StorageObjectSpec &ExternalInfoImpl::spec(size_t i) const
{
  return specs_[i];
}

size_t ExternalInfoImpl::nSpecs() const
{
  return specs_.length();
}

StorageObjectSpec::StorageObjectSpec()
: storageType(0), codingSystem(0), flags(0)
{
}

StorageObjectPosition::StorageObjectPosition()
: endOffset(Offset(-1)), line1RS(0), startsWithRS(0), insertedRSs(0)
{
}
